import unittest
from fabric.config import Config
import yaml
from pathlib import Path
from rucksack.config_parser import *
from unittest.mock import patch, mock_open

mock_config = """
---
system:
  get-uptime:
    command: uptime
"""

def mock_glob(file):
  if file == 'fake/*.yml':
    return ['file1.yml']
  if file == 'fake/*.yaml':
    return ['file2.yaml']

def load_config_from_file_cwd(file):
  return yaml.safe_load(mock_config)

def mock_find_config_files_in_directory(directory):
  if directory == f"{Path.home()}/.config/rucksack":
    return [f"{Path.home()}/.config/rucksack/file1.yml", f"{Path.home()}/.config/rucksack/file2.yaml"]
  elif directory == '/etc/rucksack':
    return ['/etc/rucksack/file1.yml', '/etc/rucksack/file2.yaml']

def mock_return_config(*args, **kwargs):
  return yaml.safe_load(mock_config)

def mock_raise_oserror(*args, **kwargs):
  raise OSError('Unable to read config file')


def mock_return_empty(*args, **kwargs):
  return {}


class TestConfigParser(unittest.TestCase):

  @patch('builtins.open', new_callable=mock_open, read_data=mock_config)
  def test_load_config_from_file(self, mockopen):
    expected_config = yaml.safe_load(mock_config)
    config = load_config_from_file('fake')
    self.assertEqual(config, expected_config)

  @patch('glob.glob', mock_glob)
  def test_find_config_files_in_directory(self):
    files = find_config_files_in_directory('fake')
    self.assertEqual(files, ['file1.yml', 'file2.yaml'])

  @patch('rucksack.config_parser.load_config_from_file', load_config_from_file_cwd)
  def test_load_config_cwd(self):
    config = load_config_from_cwd()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.find_config_files_in_directory', mock_find_config_files_in_directory)
  @patch('rucksack.config_parser.load_config_from_file', load_config_from_file_cwd)
  def test_load_config_from_home_dir(self):
    config = load_config_from_home_dir()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.find_config_files_in_directory', mock_find_config_files_in_directory)
  @patch('rucksack.config_parser.load_config_from_file', load_config_from_file_cwd)
  def test_load_config_from_etc_dir(self):
    config = load_config_from_etc_dir()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_cwd', mock_return_config)
  def test_load_config_main_cwd(self):
    config = load_config()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_cwd', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_home_dir', mock_return_config)
  def test_load_config_main_load_config_from_home_dir(self):
    config = load_config()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_cwd', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_home_dir', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_etc_dir', mock_return_config)
  def test_load_config_main_load_config_from_etc_dir(self):
    config = load_config()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_cwd', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_home_dir', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_etc_dir', mock_return_empty)
  @patch('rucksack.config_parser.load_config_from_etc_file', mock_return_config)
  def test_load_config_main_load_config_from_etc_file(self):
    config = load_config()
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_file', mock_return_config)
  def test_load_config_from_specified_file(self):
    config = load_config(file='test')
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  @patch('rucksack.config_parser.load_config_from_directory', mock_return_config)
  def test_load_config_from_specified_directory(self):
    config = load_config(directory='test')
    expected_config = yaml.safe_load(mock_config)
    self.assertEqual(config, expected_config)

  def test_validate_config(self):
    config = {
      'rucksack-config': {
        'log-level': 'debug',
        'illegal-parameter': 'fail the test'
      }
    }
    self.assertRaises(ConfigParserException, validate_config, config)

    config = {
      "validCommand": ""
    }
    validation_result = validate_config(config)
    self.assertEqual(validation_result, True)

  @patch('rucksack.config_parser.load_config', mock_return_config)
  def test_get_config_config_found(self):
    config = get_config()
    self.assertEqual(config, yaml.safe_load(mock_config))

  @patch('rucksack.config_parser.load_config', mock_return_empty)
  def test_get_config_config_not_found(self):
    self.assertRaises(ConfigNotFoundException, get_config)

  @patch('rucksack.config_parser.find_config_files_in_directory', mock_find_config_files_in_directory)
  @patch('rucksack.config_parser.load_config_from_file', mock_return_config)
  def test_load_config_from_directory(self):
    config = load_config_from_directory('/etc/rucksack')
    self.assertEqual(config, yaml.safe_load(mock_config))

    config = load_config_from_directory('/bad/directory')
    self.assertEqual(config, None)

  @patch('rucksack.config_parser.find_config_files_in_directory', mock_find_config_files_in_directory)
  @patch('rucksack.config_parser.load_config_from_file', mock_raise_oserror)
  def test_load_config_from_directory_exception(self):
    self.assertRaises(ConfigParserException, load_config_from_directory, '/etc/rucksack')



import logging
from fabric import Connection


class RucksackConnection():
  def __init__(self, host):
    self.host = host
    self.logger = logging.getLogger(__name__)
    self.logger.debug(f"Attempting to open connection with hosts: {host}")
    self.conn = Connection(host)

    # Connections are lazy, so eagerly open the connection (as long as it isn't to localhost)
    if host in ['localhost', '127.0.0.1']:
      pass
    else: 
      self.conn.open()
    self.logger.debug(f"Opened connection with host {host}")

  def execute_command(self, command):
    self.logger.debug(f"Attempting to execute command: {command}")

    if self.host in ['localhost', '127.0.0.1']:
        return self.conn.local(command, warn=True, hide=True)
    else:
      return self.conn.run(command, warn=True, hide=True)



import logging
from fabric import config
from jinja2 import Environment
from .config_parser import args_list_to_dictionary
from .exceptions import MandatoryArgumentMissingException, UnknownArgumentException, UserWantsToQuitException, ArgumentValueNotProvidedException

logger = logging.getLogger(__name__)

# Given a string and dictionary
# While the string has items: pop an item off the string
  # Can it be found in the current dictionary?
    # Yes
      # Is it a command string?
        # Yes - set the command string variable
        # No
          # Does it have argument options? (e.g., a list etc.)
            # Yes - Present these options to the user
            # No - it must be an argument. Add it to the argument array/dictionary
  # Update the current key and dictionary
def parse_command(requested_command, config):
  current_dictionary = config
  available_args = {}
  command_arguments = {}
  command_config = {}
  command_string = None

  logger.debug(f"Command to parse: {requested_command}")

  # Check to see if the user entered a "special" token
  check_special_command(requested_command)

  while True:
    try:
      current_token = requested_command.pop(0).strip()
      logger.debug(f"Current token: {current_token}")
    except IndexError:
      logger.debug('No more tokens to process')
      break

    # Check to see if we can find the current command in the current dictionary
    # If we can, then we are able to proceed further and see if this is a command,
    # an argument, a category, etc.
    if not command_string:
      logger.debug('Command string is not yet defined')
      try:
        current_dictionary = current_dictionary[current_token]
        logger.debug(f"Current dictionary set to: {current_dictionary}")
      except KeyError:
        pass

    # Check to see if the current dictionary has a "command" key.
    # If it does, then this is a command string
    try:
      if not command_string:
        logger.debug('Command string is not yet defined')
        command_config = current_dictionary
        command_string = current_dictionary['command']
        logger.debug(f"Command string found in current command dictionary: {command_string}")
        continue
    except KeyError:
      logger.debug('No command found in current command dictionary')
      pass

    # If we have found a command string, then remaining keys must be the argument
    # TODO: this doesn't account for a command flag that doesn't take an argument
    if command_string:
      try:
        available_args = args_list_to_dictionary(current_dictionary['args'])
        logger.debug(f"Found arguments for command: {available_args}")
      except KeyError:
        # TODO: Throw exception if an arg is provided but there are no args
        pass
      try:
        # First, make sure we can actually find this argument defined for the command
        # If we find it, then we pop off the next key (the actual argument) and store it
        # in the dictionary
        if current_token in available_args.keys():
          logger.debug(f"'{current_token}' is a valid argument for this command")
          try:
            argument_value = requested_command.pop(0).strip()
          except IndexError:
            logger.warning(f"No value provided for argument: {current_token}")
            raise ArgumentValueNotProvidedException(f"No value provided for argument: {current_token}")
          command_arguments[current_token] = argument_value
          logger.debug(f"Adding '{argument_value}' as the value for the '{current_token}' argument")
        else:
          logger.warning(f"Unknown argument '{current_token}'")
      except KeyError:
        raise UnknownArgumentException(f"Unkown argument: {current_token}")

  # Return three things to the calling function:
  ## 1. The command string itself, which can be passed to a rendering function
  ## 2. A dictionary of arguments, which can also be passed to a rendering function
  ## 3. The config for this entire command, which is useful for doing things like rendering
  ##    default values
  command_dictionary = {
    'command_string': command_string,
    'command_args': command_arguments,
    'command_config': command_config
  }

  logger.debug(f"Returning command dictionary: {command_dictionary}")
  return command_dictionary

def check_special_command(requested_command):
  logger.debug(f"Checking to see if 'requested_command' is a special command.")
  token = requested_command[0].strip()
  if token == "quit":
    logger.debug('"Special command found: quit')
    raise UserWantsToQuitException

def check_mandatory_args(command_dictionary):

  logger.debug(f"Checking mandatory arguments for command dictionary: {command_dictionary}")

  # Check to see if the command should have args. If not, no need to proceed
  try:
    configured_args = args_list_to_dictionary(command_dictionary['command_config']['args'])
  except KeyError:
    logger.debug('Command dictionary does not have any args, so nothing to check.')
    return

  # Iterate through all args for the command and see if any are mandatory
  # If they are, confirm they have been provided
  for configured_arg in configured_args:
    logger.debug(f"Checking to see if '{configured_arg}' is mandatory.")
    try:
      if configured_args[configured_arg]['mandatory']:
        logger.debug(f"Arg '{configured_arg}' is mandatory")
        mandatory_arg_found = False
        for provided_arg in command_dictionary['command_args']:
          if provided_arg == configured_arg:
            logger.debug(f"Found mandatory arg '{configured_arg}' in user provided args.")
            mandatory_arg_found = True

        # If a mandatory arg is not found, see if there is a default for it
        if not mandatory_arg_found:
          logger.debug(f"Mandatory arg '{configured_arg}' not found in user provided args.")
          try:
            configured_args[configured_arg]['default']
            logger.debug(f"Mandatory arg '{configured_arg}' has a default value")
          except KeyError:
            logger.warning(f"Mandatory arg '{configured_arg}' not found in user provided args and has no default value.")
            raise MandatoryArgumentMissingException(f"The mandatory arg {configured_arg} was not provided.")
    except KeyError:
      logger.debug(f"Arg '{configured_arg}' is not mandatory.")
      pass

def render_command(command_dictionary):

  check_mandatory_args(command_dictionary)
  
  template_string = command_dictionary['command_string']

  # For each command argument that is configured for the command, check to see if the user
  # provided a value for it. If so, append its template to the command.
  # We iterate over the config dict first because the order of the arguments matters.
  try:
    for arg in command_dictionary['command_config']['args']:
      logger.debug(f"Checking argument to render: {arg}")
      try:
        # The arg name is the only key in the dictionary, so obtain it.
        arg_name = list(arg)[0]

        try:
          # Get the template string if the arg was provided
          if command_dictionary['command_args'][arg_name]:
            try:
              # Check to see if the arg has a template string.
              # If not, it must be part of the command string, so continue.
              arg_template_string = arg[arg_name]['arg_string']
              logger.debug(f"Found arg template string: {arg_template_string}")
              template_string = f"{template_string} {arg_template_string}"
            except KeyError:
              continue
            continue
        except KeyError:
          # If the arg wasn't provided, then add the default to the command args dictionary
          logger.debug(f"Arg not provided: {arg_name}")
          if arg[arg_name]['mandatory']:
            arg_value = arg[arg_name]['default']
            command_dictionary['command_args'][arg_name] = arg_value
            logger.debug(f"Found default value of '{arg_value}' for arg '{arg_name}")
      except KeyError:
        pass
  except KeyError:
    # If there are no args, then there's nothing to do!
    logger.debug(f"No args for command, nothing to do.")
    pass

  env = Environment()
  template = env.from_string(template_string)
  logger.debug(f"Rendering Jinja2 template: {template_string}")
  return template.render(command_dictionary['command_args'])



import unittest
from rucksack.command_parser import *
from rucksack.exceptions import MandatoryArgumentMissingException

class TestCommandParser(unittest.TestCase):
  def test_parse_command(self):
    config = {
      'system': {
        'tail-log': {
          'command': 'tail',
          'args': [
            {'log_file': {}},
            {'num_lines': {}}
          ]
        }
      }
    }

    expected_result = {
      'command_string': 'tail',
      'command_args': {'log_file': '/var/log/messages', 'num_lines': '10'},
      'command_config': {
        'command': 'tail',
        'args': [
          {'log_file': {}},
          {'num_lines': {}}
        ]
      }
    }

    command_string = 'system tail-log log_file /var/log/messages num_lines 10'
    self.assertEqual(expected_result,parse_command(command_string.split(), config)
)

  def test_check_mandatory_args(self):
    # Test with mandatory arg missing
    command_dictionary = {'command_string': 'tail', 'command_args': {}, 'command_config': {'command': 'tail', 'args': [{'num_lines': {'arg_string': '-n {{ num_lines }}'}}, {'log_file': {'mandatory': True}}]}}
    self.assertRaises(MandatoryArgumentMissingException, check_mandatory_args, command_dictionary)

    # Test with mandatory arg provided
    command_dictionary = {'command_string': 'tail', 'command_args': {'log_file': '/var/log/messages'}, 'command_config': {'command': 'tail', 'args': [{'num_lines': {'arg_string': '-n {{ num_lines }}'}}, {'log_file': {'mandatory': True}}]}}
    check_mandatory_args(command_dictionary)

    # Test with no args
    command_dictionary = {'command_string': 'uptime', 'command_args': {}, 'command_config': {'command': 'uptime'}}
    check_mandatory_args(command_dictionary)

  def test_check_special_command(self):
    command = ['quit', 'whatever']
    self.assertRaises(UserWantsToQuitException, check_special_command, command)

  def test_render_command(self):

    command_dictionary = {
      'command_string': 'tail {{ log_file }}',
      'command_args': {
        'num_lines': '5'
        },
        'command_config': {
          'command': 'tail {{ log_file }}',
          'args': [
            {
              'num_lines': {
                'arg_string': '-n {{ num_lines }}'
                }
            },
            {
              'log_file': {
                'mandatory': True,
                'default': '/var/log/syslog',
                'values': ['/var/log/syslog', '/var/log/kern.log', '/var/log/auth.log']
                }
            }
          ]
        }
      }

    expected_result = 'tail /var/log/syslog -n 5'

    self.assertEqual(render_command(command_dictionary), expected_result)

  def test_render_command_no_args(self):
    command_dictionary = {'command_string': 'uptime', 'command_args': {}, 'command_config': {'command': 'uptime'}}
    expected_result = 'uptime'
    self.assertEqual(render_command(command_dictionary), expected_result)






import unittest
from unittest.mock import patch
from rucksack.connection import *

class MockConnection():
  def __init__(self, host):
    pass

  def open(host):
    pass

  def run(self, *args, **kwargs):
    return True

  def local(self, *args, **kwargs):
    return True

@patch('rucksack.connection.Connection', MockConnection)
class TestToolbagConnection(unittest.TestCase):

  def test_connection(self):
    my_connection = RucksackConnection("www.example.com")
    command_result = my_connection.execute_command("whoami")

    self.assertEqual(command_result, True)

@patch('rucksack.connection.Connection', MockConnection)
class TestToolbagLocalConnection(unittest.TestCase):

  def test_connection(self):
    my_connection = RucksackConnection("localhost")
    command_result = my_connection.execute_command("whoami")

    self.assertEqual(command_result, True)



import logging
import sys
from prompt_toolkit import PromptSession, print_formatted_text
from prompt_toolkit.completion import FuzzyCompleter
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style
from rucksack.completer import RucksackCompleter
from rucksack.command_parser import parse_command, render_command
from rucksack.connection import RucksackConnection
from rucksack.config_parser import get_config
from rucksack.exceptions import ArgumentValueNotProvidedException, ConfigNotFoundException, UnknownArgumentException, UserWantsToQuitException, MandatoryArgumentMissingException


class RucksackCli():

  def __init__(self, host, config_file=None, config_directory=None):
    self.host = host
    self.logger = logging.getLogger(__name__)

    self.logger.debug('Creating new RucksackCLI')

    self.logger.debug('Attempting to get configuration')
    self.config = get_config(config_file, config_directory)

    self.logger.debug(f"Attempting to create a RucksackConnection to {host}")
    self.conn = RucksackConnection(self.host)


  # Sets the color of the command prompt based on whether or not there was an error
  # Command prompt is red on error, white otherwise
  # TODO: would be cool to customize this via config
  def set_prompt(self, command_error):

    if command_error:
      parser_color = 'red'
    else:
      parser_color = 'white'

    style = Style.from_dict({
        'prompt': parser_color,
    })

    message = [
        ('class:prompt', f"{self.host}> ")
    ]

    return {
      "message": message,
      "style": style
    }

  def start(self):

    session = PromptSession()

    command_error = None

    while True:

      text = session.prompt(**self.set_prompt(command_error), completer=FuzzyCompleter(RucksackCompleter(self.config, self.conn)), complete_in_thread=True)

      requested_command = text.split()
      self.logger.debug(f"User entered requested command f{requested_command}")

      if not requested_command:
        self.logger.debug('User did not enter a command')
        continue

      try:
        command_string = parse_command(requested_command, self.config)
        self.logger.debug(f"Command parsed as {command_string}")
      except (UnknownArgumentException, ArgumentValueNotProvidedException) as e:
          self.logger.debug('Command parsing failed')
          text = FormattedText([
              ('red', str(e)),
          ])
          print_formatted_text(text, file=sys.stderr)
          continue

      if command_string['command_string']:
        try:
          rendered_command = render_command(command_string)
          self.logger.debug(f"Command rendered as: {rendered_command}")
        except MandatoryArgumentMissingException as e:
          text = FormattedText([
              ('red', str(e)),
          ])
          print_formatted_text(text, file=sys.stderr)
          continue
      else:
        text = FormattedText([
            ('red', 'No such command'),
        ])

        print_formatted_text(text, file=sys.stderr)
        continue
      self.logger.info(f"Attempting to run: {rendered_command}")
      print_formatted_text(f"Attempting to run {rendered_command}", file=sys.stderr)
      result = self.conn.execute_command(rendered_command)

      self.logger.debug(f"Command result: {result}")

      if result.stderr:
        command_error = True
      else:
        command_error = False

      stdout = FormattedText([
        ('green', result.stdout)
      ])

      stderr = FormattedText([
        ('red', result.stderr)
      ])

      print_formatted_text(stdout)
      print_formatted_text(stderr, file=sys.stderr)



import unittest
import yaml
from rucksack.completer import *
from prompt_toolkit.document import Document
from invoke.runners import Result


mock_config = """
---
rucksack-config:
  somevalue: 'test'
system:
  get-uptime:
    command: "uptime"
  view-auth-log:
    command: "cat"
    args:
      - log:
          from_command: "find /var/log -name '*auth.log*' | grep -v '.gz'"
  tail-log:
    command: "tail"
    args:
      - num_lines:
         arg_string: -n {{ num_lines }}
      - log_file:
          mandatory: True
          values:
            - /var/log/messages
            - /var/log/kern.log
            - /var/log/auth.log
apache:
  get-ips:
    command: "tail /var/log/apache2/access_log"
    args:
      - num_lines:
          arg_string: -n {{ num_lines }}
nginx:
  reload:
    command: 'kill -HUP $(pidof nginx)'
"""

class MockConnection():
  def execute_command(self, command):
    return Result(stdout='/var/log/auth.log')

class TestCompleter(unittest.TestCase):

  def test_completer_with_section(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("system")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, "get-uptime")

  def test_completion_with_values(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("system tail-log log_file")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, "/var/log/messages")

  def test_completion_with_from_commmand(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), MockConnection())
    document = Document("system view-auth-log log")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, "/var/log/auth.log")

  def test_completion_with_nothing(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, "system")

  def test_completion_with_bad_arg(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("system tail-log somearg ")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertIn(yielded_completion.text, ['num_lines', 'log_file'])

  def test_completion_with_bad_first_term(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("test")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, '')

  def test_completion_with_user_typing_arg_value(self):
    completer = RucksackCompleter(yaml.load(mock_config, Loader=yaml.Loader), None)
    document = Document("system tail-log log_file /var/lo")
    yielded_completion = next(completer.get_completions(document, None))
    self.assertEqual(yielded_completion.text, '')



class ConfigParserException(Exception):
  pass

class MandatoryArgumentMissingException(Exception):
  pass

class UserWantsToQuitException(Exception):
  pass

class UnknownArgumentException(Exception):
  pass

class ArgumentValueNotProvidedException(Exception):
  pass

class ConfigNotFoundException(Exception):
  pass



import unittest
from prompt_toolkit.formatted_text import FormattedText
from unittest.mock import patch
from rucksack.cli import *

config = {
  'uptime': {
    'command': 'uptime'
  },
  'tail-log': {
    'command': 'tail {{ logfile }}',
    'args': [
      { 'logfile': {'mandatory': True} }
    ]
  },
  'gimme-an-error': {
    'command': 'false'
  }
}

# I know, I know: global variables
# The command array array allows the MockPromptSession to pop off commands provided by tests
# The output array stores output from all mocked command executions, so it can be compared to
# expected output
commands = []
output = []

# Class to mock session. Primarily used to mock sending a user command by popping a command
# off the commands array
class MockPromptSession():

  def prompt(self, *args, **kwargs):
    return commands.pop(0)

# Simple mock for command output
class MockCommandOutput():
  def __init__(self, stdout, stderr):
    self.stdout = stdout
    self.stderr = stderr

# TODO: this is copy/pasted from test_connection
class MockConnection():
  def __init__(self, host):
    pass

  def local(self, command, **kwargs):
    if command == 'false':
      return MockCommandOutput('', 'bad stuff')
    else:
      return MockCommandOutput('mockoutput', '')

def mock_get_config(*args, **kwargs):
  return config

def mock_get_config_with_error(*args, **kwargs):
  raise ConfigNotFoundException('Configuration not found')

# Throws stdout/stderr into a variable that can be examined later
def mock_print_formatted_text(text, **kwargs):
  global output
  output.append(text)
class TestCLI(unittest.TestCase):

  @patch('rucksack.cli.get_config', mock_get_config)
  @patch('rucksack.cli.PromptSession', MockPromptSession)
  @patch('rucksack.connection.Connection', MockConnection)
  @patch('rucksack.cli.print_formatted_text', mock_print_formatted_text)
  def run_cli_test(self, commands_to_execute, expected_output):
    global commands
    global output
    output = []
    commands = commands_to_execute
    ruck_cli = RucksackCli('localhost')
    self.assertRaises(UserWantsToQuitException, ruck_cli.start)
    self.assertEqual(output, expected_output)

  def test_valid_command(self):
    commands_to_execute = ['uptime', 'quit']
    expected_output = ['Attempting to run uptime', FormattedText([('green', 'mockoutput')]), FormattedText([('red', '')])]
    self.run_cli_test(commands_to_execute, expected_output)

  def test_no_command(self):
    commands_to_execute = ['', 'quit']
    expected_output = []
    self.run_cli_test(commands_to_execute, expected_output)

  def test_bad_command(self):
    commands_to_execute = ['wrongcommand', 'quit']
    expected_output = [FormattedText([('red', 'No such command')])]
    self.run_cli_test(commands_to_execute, expected_output)

  def test_missing_mandatory_arg(self):
    commands_to_execute = ['tail-log', 'quit']
    expected_output = [FormattedText([('red', 'The mandatory arg logfile was not provided.')])]
    self.run_cli_test(commands_to_execute, expected_output)

  def test_missing_arg_value(self):
    commands_to_execute = ['tail-log logfile', 'quit']
    expected_output = [FormattedText([('red', 'No value provided for argument: logfile')])]
    self.run_cli_test(commands_to_execute, expected_output)

  def test_command_with_error(self):
    commands_to_execute = ['gimme-an-error', 'quit']
    expected_output = ['Attempting to run false', FormattedText([('green', '')]), FormattedText([('red', 'bad stuff')])]
    self.run_cli_test(commands_to_execute, expected_output)

  @patch('rucksack.cli.get_config', mock_get_config_with_error)
  def test_config_not_found(self):
    self.assertRaises(ConfigNotFoundException, RucksackCli, 'localhost')



import glob
import logging
import os
import yaml
from pathlib import Path
from .exceptions import ConfigParserException, ConfigNotFoundException

logger = logging.getLogger(__name__)

def args_list_to_dictionary(args_list):
  logger.debug(f"Converting arg list to dictionary: {args_list}")
  args_dictionary = {}
  for item in args_list:
    arg_name = list(item.keys())[0]
    args_dictionary[arg_name] = item[arg_name]
  logger.debug(f"Returning arg dictionary: {args_dictionary}")
  return args_dictionary

def load_config_from_file(file=None):
  logger.debug(f"Attempting to load config from file: {file}")
  with open(file, 'r') as config_file:
    config = yaml.load(config_file, Loader=yaml.Loader)
    logger.debug(f"Config loaded from file. Returning config: {config}")
    return config

# TODO: utilize this function elsewhere and DRY out code
def load_config_from_directory(directory):
  logger.debug(f"Attempting to load config file from directory: {directory}")
  config = {}
  files = find_config_files_in_directory(directory)
  if files:
    logger.debug(f"Found config files in directory: {directory}")
    for file in files:
      try:
        temp_dict = load_config_from_file(file)
        for key in temp_dict.keys():
          config[key] = temp_dict[key]
      except OSError as e:
        logger.debug(f"Unable to read config file {file}")
        raise ConfigParserException(f"Unable to read config file {file}") from e
    return config
  else:
    logger.debug(f"No config files found in {directory}") 

def find_config_files_in_directory(directory=None):
  logger.debug(f"Attempting to find config files in directory {directory}")
  files = []
  files += glob.glob(f"{directory}/*.yml")
  files += glob.glob(f"{directory}/*.yaml")
  logger.debug(f"Config files in directory: {files}")
  return files

def load_config_from_cwd():
  logger.debug('Attempting to load config file from current working directory')
  config = {}
  file = f"{os.getcwd()}/rucksack.yml"
  try:
    config = load_config_from_file(file)
    logger.debug('Successfully loaded config file from CWD rucksack.yml')
  except OSError:
    logger.debug(f"Unable to read config file {file}")
    pass

  file = f"{os.getcwd()}/rucksack.yaml"
  try:
    config = load_config_from_file(file)
    logger.debug('Successfully loaded config file from CWD rucksack.yaml')
  except OSError:
    logger.debug(f"Unable to read config file {file}")
    pass

  return config

def load_config_from_home_dir():
  logger.debug('Attempting to load config file from home directory')
  config = {}
  files = find_config_files_in_directory(f"{Path.home()}/.config/rucksack")
  if files:
    logger.debug(f"Found config files in home {Path.home()}/.config/rucksack")
    for file in files:
      try:
        temp_dict = load_config_from_file(file)
        for key in temp_dict.keys():
          config[key] = temp_dict[key]
      except OSError as e:
        logger.debug(f"Unable to read config file {file}")
        raise ConfigParserException(f"Unable to read config file {file}") from e
    return config
  else:
    logger.debug(f"No config files found in {Path.home()}/.config/rucksack")

def load_config_from_etc_dir():
  logger.debug('Attempting to load config file from /etc/rucksack directory')
  config = {}
  files = find_config_files_in_directory('/etc/rucksack')
  if files:
    logger.debug('Found config files in /etc/rucksack')
    for file in files:
      try:
        temp_dict = load_config_from_file(file)
        for key in temp_dict.keys():
          config[key] = temp_dict[key]
      except OSError as e:
        logger.critical(f"Unable to read config file {file}")
        raise ConfigParserException(f"Unable to read config file {file}") from e
    return config
  else:
    logger.debug(f"No config files found in /etc/rucksack")

def load_config_from_etc_file():
  config = {}
  try:
    config = load_config_from_file(f"{Path.home()}/.config/rucksack/rucksack.yml")
  except OSError:
    logger.debug(f"Unable to read config file {Path.home()}/.config/rucksack/rucksack.yml")
    pass

  try:
    config = load_config_from_file(f"{Path.home()}/.config/rucksack/rucksack.yaml")
  except OSError:
    logger.debug(f"Unable to read config file {Path.home()}/.config/rucksack/rucksack.yaml")
    pass

  return config

# Config file search order
## Skipped if the user passes a config directory or file
## Any match halts the search process)
### 1. Look for rucksack.[yml|yaml] in local directory
### 2. Search ~/.config/rucksack for yaml or yml files. Load all files.
### 3. Look for ~/.config/rucksack.[yml|yaml]
### 4. Search /etc/rucksack for yaml or yml files. Load all files.
def load_config(file=None, directory=None):
  config = {}

  # Attempt to load from a file or directory if provided
  if file:
    config = load_config_from_file(file)
    return config
  if directory:
    config = load_config_from_directory(directory)
    return config
  

  config = load_config_from_cwd()
  if config:
    return config

  config = load_config_from_home_dir()
  if config:
    return config

  config = load_config_from_etc_dir()
  if config:
    return config

  config = load_config_from_etc_file()
  if config:
    return config

def validate_config(config):

  if 'rucksack-config' in config.keys():
    validate_rucksack_config(config['rucksack-config'])

  # TODO: more validation would be great (e.g., ensuring a command only has valid keys, etc.)
  return True

def validate_rucksack_config(config):
  # TODO: these are CLI args, so eliminate them from config
  allowed_keys = {'log-level', 'log-file'}
  illegal_keys = set(config.keys()) - allowed_keys
  if illegal_keys:
    raise ConfigParserException(f"Illegal keys found in rucksack-config: {illegal_keys}")

def get_config(file=None, directory=None):
  config = load_config(file, directory)
  if not config:
    raise ConfigNotFoundException("No valid configuration file found.")
  validate_config(config)
  return config



import logging
from prompt_toolkit.completion import Completer, Completion
from functools import lru_cache
from .config_parser import args_list_to_dictionary

logger = logging.getLogger(__name__)

@lru_cache
def get_args_from_command(connection, command):
  # TODO: handle errors
  logger.debug(f"Attempting to obtain arguments from command: {command}")
  result = connection.execute_command(command)
  logger.debug(f"Command result: {result}")
  return result.stdout.splitlines()

class RucksackCompleter(Completer):

  def __init__(self, config, connection):
    self.logger = logging.getLogger(__name__)
    self.config = config
    self.connection = connection

  def get_completions(self, document, complete_event):
    args_dictionary = None
    current_args = set()
    requested_command = document.text.split()
    self.logger.debug(f"Attempting to find completions for: {requested_command}")

    # TODO: consider tracking more state at the class level to reduce duplication.
    # Would need to reset the args dictionary if we exist the args
    # Can do this by storing the position of where we enter the args

    current_dictionary = self.config
    values_to_yield = self.config.keys()
    while True:
      try:
        current_term = requested_command.pop(0).strip()
        self.logger.debug(f"Processing current term: {current_term}")
      except IndexError:
        self.logger.debug(f"No more terms to process in requested command")
        break

      # If we have an args_dictionary, then we've already found the command. No need to keep
      # trying to process the config dict for subcommmands
      if args_dictionary:
        self.logger.debug(f"Args dictionary is currently set: {args_dictionary}")
        self.logger.debug('Checking to see if args dictionary contains current term')
        # If the current term is in the args dictionary and has values, present those to the user
        try:
          # Add this term to a set of current args so that we can avoid presenting
          # duplicate args to the user
          current_dictionary = args_dictionary[current_term]
          self.logger.debug('Term found in args dictionary. Adding to current args set')
          current_args.add(current_term)
          # Ensure that current values to yield (which may be set from previous reference to args dict keys)
          # has any current args removed
          values_to_yield = set(values_to_yield) - current_args
        except KeyError:
          # If the current term isn't in the args dictionary, then default to giving the user
          # the list of args
          self.logger.debug('Term not found in args dictionary.')
          if document.char_before_cursor == ' ':
            self.logger.debug('Character before cursor is a space, yielding all args not yet provided.')
            values_to_yield = set(args_dictionary.keys()) - current_args
            continue
          else:
            self.logger.debug('Non-space character before cursor, user must be typing an arg value.')
            yield Completion("", start_position=0)
      else:
        # If we don't have an args_dictionary, then we haven't found the command yet. Keep
        # diving into the dictionary in an effort to find it.
        self.logger.debug('Args dictionary is not yet set. Continuing to search for command.')
        try:
          current_dictionary = current_dictionary[current_term]
          values_to_yield = current_dictionary.keys()
          self.logger.debug(f"Found current term in dictionary. Yielding: {values_to_yield}")
        except KeyError:
          # The user must've given us some invalid data.
          yield Completion("", start_position=0)

      # Check to see if this current dictionary has a command key
      # If it does, we're inside a command and everything else is the args
      try:
        if not args_dictionary and current_dictionary['command']:
          args_dictionary = args_list_to_dictionary(current_dictionary['args'])
          values_to_yield = set(args_dictionary) - current_args
          self.logger.debug(f"Current dictionary has a command string. Yielding args: {args_dictionary}")
      except KeyError:
        # A command may not have args
        self.logger.debug('Command found, but current dictionary does not have args.')
        pass

      # If there is a values field in the current dictionary, then present these to the user
      try:
        values_to_yield = current_dictionary['values']
        self.logger.debug(f"Current dictionary has values. Yielding: {values_to_yield}")
      except KeyError:
        pass

      # If there is a from_command field in the current dictionary, then obtain the appropriate set of values
      try:
        command_to_execute = current_dictionary['from_command']
        self.logger.debug('Current dictionary has from_command. Attempting to obtain args from command.')
        values_to_yield = get_args_from_command(self.connection, command_to_execute)
        self.logger.debug(f"Received args from command: {values_to_yield}")
      except KeyError:
        pass

    if not values_to_yield and args_dictionary:
      self.logger.debug('No values to yield. Yielding keys of arg dictionary.')
      values_to_yield = set(args_dictionary.keys()) - current_args

    for value in values_to_yield:
      if value == 'command' or value == 'rucksack-config':
        continue
      yield Completion(value, start_position=0)
